# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
from reahl.tofu import Fixture, set_up, temp_dir
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import stubclass, replaced

from reahl.webdev.tools import Browser

from reahl.component.context import ExecutionContext
from reahl.web.dhtml import DhtmlUI, DHTMLFile
from reahl.web.fw import UserInterface
from reahl.web.ui import HTML5Page

from reahl.web_dev.fixtures import WebFixture, BasicPageLayout
from reahl.component_dev.test_i18n import LocaleContextStub


class DhtmlFixture(Fixture):
    def new_static_dir(self):
        return temp_dir()
    
    div_internals = '<p>some stúff</p>'

    def new_dhtml_file(self):
        charset = 'UTF-8'
        contents = '<!DOCTYPE html><meta charset="%s"><html><head><title>â title</title></head>' \
                   '<body><div id="astatic">%s</div></body></html>' % (charset, self.div_internals)
        return self.static_dir.file_with('correctfile.d.html', contents.encode(charset), mode='w+b')

    def new_afrikaans_dhtml_file(self):
        contents = '<!DOCTYPE html><meta charset="UTF-8"><html><head><title>Afrikaans bo!</title></head><body></body></html>'
        return self.static_dir.file_with('correctfile.af.d.html', contents)

    def new_other_file(self):
        return self.static_dir.file_with('otherfile.txt', 'other')

    def get_inserted_html(self, browser):
        return browser.get_html_for('//div[contains(@class, "column-main")]/*')

    @set_up
    def create_files(self):
        self.dhtml_file
        self.afrikaans_dhtml_file
        self.other_file


@with_fixtures(WebFixture, DhtmlFixture)
def test_basic_workings(web_fixture, dhtml_fixture):
    """A DhtmlUI provides a UserInterface which maps to the filesystem where there may be
       a combination of .d.html and other files. When a d.html file is requested from
       it, the contents of the specified div from inside the d.html file is inserted
       into the specified Slot. When a normal file is requested, the file is sent verbatim."""

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/dhtml_ui', DhtmlUI, {'main_slot': 'main'},
                            name='test_ui', static_div_name='astatic')

    fixture = dhtml_fixture

    # Dhtml files should be located in the web.static_root
    web_fixture.config.web.static_root = fixture.static_dir.name

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI, enable_js=True)
    browser = Browser(wsgi_app)

    # A dhtml file: HTML5Page's main_slot now contains the insides of the div in the dhtml file
    browser.open('/dhtml_ui/correctfile.d.html')
    html = fixture.get_inserted_html(browser)
    assert html == fixture.div_internals

    # A non-dhtml file is returned verbatim
    browser.open('/dhtml_ui/otherfile.txt')
    contents = browser.raw_html
    assert contents == 'other'

    # Files that do not exist are reported as such
    browser.open('/dhtml_ui/idonotexist.txt', status=404)
    browser.open('/dhtml_ui/idonotexist.d.html', status=404)


@with_fixtures(WebFixture, DhtmlFixture)
def test_i18n_dhtml(web_fixture, dhtml_fixture):
    """Dhtml files can have i18nsed versions, which would be served up if applicable."""

    class MainUI(UserInterface):
        def assemble(self):
            self.define_page(HTML5Page).use_layout(BasicPageLayout())
            self.define_user_interface('/dhtml_ui', DhtmlUI, {'main_slot': 'main'},
                                       name='test_ui', static_div_name='astatic')

    fixture = dhtml_fixture

    # Dhtml files should be located in the web.static_root
    web_fixture.config.web.static_root = fixture.static_dir.name

    wsgi_app = web_fixture.new_wsgi_app(site_root=MainUI)

    browser = Browser(wsgi_app)

    # request the file, but get the translated alternative for the locale
    def stubbed_create_context_for_request():
        return LocaleContextStub(locale='af')
    with replaced(wsgi_app.create_context_for_request, stubbed_create_context_for_request):
        browser.open('/dhtml_ui/correctfile.d.html')

    assert browser.title == 'Afrikaans bo!'


@with_fixtures(DhtmlFixture)
def test_encoding_dammit(dhtml_fixture):
    """ """
    dhtml_file = DHTMLFile(dhtml_fixture.dhtml_file.name, ['astatic'])
    dhtml_file.read()
    assert dhtml_file.title == 'â title'
    assert dhtml_file.elements == {'astatic': dhtml_fixture.div_internals}
    assert dhtml_file.original_encoding == 'utf-8'



from setuptools import setup, Command
class InstallTestDependencies(Command):
    user_options = []
    def run(self):
        import sys
        import subprocess
        if self.distribution.tests_require: subprocess.check_call([sys.executable, "-m", "pip", "install", "-q"]+self.distribution.tests_require)

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

setup(
    name='reahl-web',
    version='4.0.2',
    description='The core Reahl web framework',
    long_description='Reahl is a web application framework that allows a Python programmer to work in terms of useful abstractions - using a single programming language.\n\nThis package contains the core of the Reahl framework.\n\nSee http://www.reahl.org/docs/4.0/tutorial/gettingstarted-install.d.html for installation instructions. ',
    url='http://www.reahl.org',
    maintainer='Iwan Vosloo',
    maintainer_email='iwan@reahl.org',
    packages=['reahl', 'reahl.messages', 'reahl.web', 'reahl.web.bootstrap', 'reahl.web.holder', 'reahl.web.static', 'reahl.web.static.jquery', 'reahl.web_dev', 'reahl.web_dev.advanced', 'reahl.web_dev.advanced.subresources', 'reahl.web_dev.appstructure', 'reahl.web_dev.bootstrap', 'reahl.web_dev.inputandvalidation', 'reahl.web_dev.widgets'],
    py_modules=[],
    include_package_data=True,
    package_data={'': ['*/LC_MESSAGES/*.mo']},
    namespace_packages=['reahl', 'reahl.messages'],
    install_requires=['reahl-component>=4.0,<4.1', 'ply>=3.8,<3.8.999', 'slimit>=0.8,<0.8.999', 'cssmin>=0.2,<0.2.999', 'BeautifulSoup4>=4.6,<4.6.999', 'webob>=1.4,<1.4.999', 'Babel>=2.1,<2.1.999', 'setuptools>=32.3.1', 'lxml>=4.2,<4.2.999'],
    setup_requires=['setuptools-git>=1.1,<1.1.999', 'pytest-runner'],
    tests_require=['pytest>=3.0', 'reahl-tofu>=4.0,<4.1', 'reahl-stubble>=4.0,<4.1', 'reahl-sqlalchemysupport>=4.0,<4.1', 'reahl-postgresqlsupport>=4.0,<4.1', 'reahl-web-declarative>=4.0,<4.1', 'reahl-domain>=4.0,<4.1', 'reahl-webdev>=4.0,<4.1', 'reahl-mailutil>=4.0,<4.1', 'reahl-dev>=4.0,<4.1'],
    test_suite='reahl.web_dev',
    entry_points={
        'reahl.eggs': [
            'Egg = reahl.component.eggs:ReahlEgg'    ],
        'reahl.translations': [
            'reahl-web = reahl.messages'    ],
        'reahl.configspec': [
            'config = reahl.web.egg:WebConfig'    ],
                 },
    extras_require={},
    cmdclass={'install_test_dependencies': InstallTestDependencies}
)



# Copyright 2016, 2017, 2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import Fixture
from reahl.tofu.pytestsupport import with_fixtures

from reahl.webdev.tools import XPath

from reahl.component.modelinterface import exposed, Field

from reahl.web.bootstrap.ui import P
from reahl.web.bootstrap.forms import Form, FormLayout, CueInput, TextInput

from reahl.web_dev.fixtures import WebFixture


class CueInputFixture(Fixture):
    cue_element_xpath = "//p"

    def new_domain_object(self):
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.field = Field(label='MyField')
        return DomainObject()


@with_fixtures(WebFixture, CueInputFixture)
def test_cue_input_display_basics(web_fixture, cue_input_fixture):
    """A CueInput displays a given cue when its wrapped Input has focus and hides the cue otherwise."""
    fixture = cue_input_fixture

    class FormWithCueInput(Form):
            def __init__(self, view):
                super(FormWithCueInput, self).__init__(view, 'test')
                self.use_layout(FormLayout())
                cue_input = CueInput(TextInput(self, fixture.domain_object.fields.field), P(view, 'this is your cue'))
                self.layout.add_input(cue_input)


    wsgi_app = web_fixture.new_wsgi_app(child_factory=FormWithCueInput.factory(), enable_js=True)

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')

    #initially the cue is not visible
    browser.wait_for_element_not_visible(fixture.cue_element_xpath)
    #tabbing to the input reveals the cue
    browser.focus_on(XPath.input_labelled('MyField'))
    browser.wait_for_element_visible(fixture.cue_element_xpath)

    #moving focus to another input causes the cue to be hidden
    browser.press_tab()
    browser.wait_for_element_not_visible(fixture.cue_element_xpath)


@with_fixtures(WebFixture, CueInputFixture)
def test_cue_is_visible_when_js_disabled(web_fixture, cue_input_fixture):
    """A CueInput degrades without JS to always display its cue."""
    fixture = cue_input_fixture

    class FormWithCueInput(Form):
            def __init__(self, view):
                super(FormWithCueInput, self).__init__(view, 'test')
                self.use_layout(FormLayout())
                cue_input = CueInput(TextInput(self, fixture.domain_object.fields.field), P(view, 'this is your cue'))
                self.layout.add_input(cue_input)


    wsgi_app = web_fixture.new_wsgi_app(child_factory=FormWithCueInput.factory(), enable_js=False)

    web_fixture.reahl_server.set_app(wsgi_app)
    browser = web_fixture.driver_browser
    browser.open('/')
    browser.refresh() # To prevent flipper we don't understand

    #the cue is visible when JS is disbled
    browser.wait_for_element_visible(fixture.cue_element_xpath)



# Copyright 2014-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
""".. versionadded:: 3.2

Tables are Widgets that show data items in rows arranged into aligned columns.
This module includes tools to help transform data for display in a Table.

You can also display a Widget (such as a Button or Input) in a Table cell by using
DynamicColumn.

The number of data items displayed in a Table can potentially be very large. DataTable
solves this problem by only displaying a small portion of the data, allowing a user
to page to different pages, each containing a suitably-sized smaller list. DataTable
can also be set up to let a user sort data according to different columns.

"""

from __future__ import print_function, unicode_literals, absolute_import, division

import six

import functools

from reahl.component.exceptions import ProgrammerError
from reahl.component.modelinterface import exposed, IntegerField, BooleanField
from reahl.web.fw import Bookmark, Widget, Layout

from reahl.web.ui import HTMLAttributeValueOption, StaticColumn, ColGroup, Col, Th, Tr, Td, Tbody, Thead, Tfoot, DynamicColumn, Caption

import reahl.web.ui
from reahl.web.bootstrap.ui import A, Div
from reahl.web.bootstrap.pagination import PageMenu, PagedPanel, SequentialPageIndex
from reahl.web.bootstrap.grid import DeviceClass


class Table(reahl.web.ui.HTMLWidget):
    """Tabular data displayed as rows broken into columns.

       :param view: (See :class:`~reahl.web.fw.Widget`)
       :keyword caption_text: If text is given here, a caption will be added to the table containing the caption text.
       :keyword summary:  A textual summary of the contents of the table which is not displayed visually, \
                but may be used by a user agent for accessibility purposes.
       :keyword css_id: (See :class:`~reahl.web.ui.HTMLElement`)

    """
    def __init__(self, view, caption_text=None, summary=None, css_id=None):
        super(Table, self).__init__(view)
        self.main_div = self.add_child(Div(view, css_id=css_id))
        self.set_html_representation(self.main_div)
        self.table = self.main_div.add_child(reahl.web.ui.Table(view, caption_text=caption_text, summary=summary))
        self.table.append_class('table')

    def with_data(self, columns, items):
        self.table.with_data(columns, items)
        return self

    @property
    def thead(self):
        return self.table.thead


class HeadingTheme(HTMLAttributeValueOption):
    def __init__(self, name):
        super(HeadingTheme, self).__init__(name, name is not None, constrain_value_to=['light', 'dark'])


class TableLayout(Layout):
    """A Layout for customising details of how a Table is displayed.

    :keyword dark: If True, table text is light text on dark background.
    :keyword border: If set to 'cells', a border is rendered around each cell.
                     When set to 'rows' (the default) only row delimeters are shown.
                     If 'borderless', no borders are rendered.
                     Allowed values: 'rows', 'cells' and 'borderless'.
    :keyword compact: If True, make the table more compact by cutting cell padding in half.
    :keyword striped: If True, colour successive rows lighter and darker.
    :keyword highlight_hovered: If True, a row is highlighted when the mouse hovers over it.
    :keyword responsive: If True, activate horizontal scrolling.  If set to a device class activate
                         horizontal scrolling only for devices smaller than the device class.
                         Allowed values : 'xs', 'sm', 'md', 'lg' and 'xl'
    :keyword heading_theme: One of 'light' or 'dark'. A light heading is one with darker text on a lighter background.
    """
    def __init__(self, dark=False, border='rows', compact=False, striped=False, highlight_hovered=False,
                 responsive=False, heading_theme=None):
        super(TableLayout, self).__init__()

        if isinstance(responsive, six.string_types):
            self.responsive_attribute_option = HTMLAttributeValueOption(responsive, True,
                                                                   prefix='table-responsive',
                                                                   constrain_value_to=DeviceClass.device_classes)
        else:
            self.responsive_attribute_option = HTMLAttributeValueOption('table-responsive', responsive)

        border_option = HTMLAttributeValueOption(border, border and border != 'rows',
                                                 prefix='table',
                                                 map_values_using={'cells': 'bordered', 'rows':''},
                                                 constrain_value_to=['borderless', 'cells', 'rows'])

        self.table_properties = [HTMLAttributeValueOption('dark', dark, prefix='table'),
                                 HTMLAttributeValueOption('striped', striped, prefix='table'),
                                 border_option,
                                 HTMLAttributeValueOption('hover', highlight_hovered, prefix='table'),
                                 HTMLAttributeValueOption('sm', compact, prefix='table')
                                 ]

        self.heading_theme = HeadingTheme(heading_theme)


    def customise_widget(self):
        super(TableLayout, self).customise_widget()

        if self.responsive_attribute_option.is_set:
            self.widget.main_div.append_class(self.responsive_attribute_option.as_html_snippet())

        for table_property in self.table_properties:
            if table_property.is_set:
                self.widget.table.append_class(table_property.as_html_snippet())
        self.style_heading()

    def style_heading(self):
        if self.heading_theme.is_set:
            table_header = self.widget.table.get_or_create_header()
            table_header.append_class('thead-%s' % self.heading_theme.as_html_snippet())


class TablePageIndex(SequentialPageIndex):
    def __init__(self, columns, items, items_per_page=10, current_page_number=1, start_page_number=1, max_page_links=5):
        super(TablePageIndex, self).__init__(items, items_per_page=items_per_page,
                                             current_page_number=current_page_number,
                                             start_page_number=start_page_number,
                                             max_page_links=max_page_links)
        self.sort_column_number = None
        self.sort_descending = False
        self.columns = columns

    @property
    def sorting_keys(self):
        return [column.sort_key for column in self.columns]

    def get_contents_for_page(self, page_number):
        if self.sort_column_number is not None:
            sorting_key = self.sorting_keys[self.sort_column_number]
            self.items.sort(key=sorting_key, reverse=self.sort_descending)

        return super(TablePageIndex, self).get_contents_for_page(page_number)

    @exposed
    def fields(self, fields):
        fields.update_copies(super(TablePageIndex, self).fields)
        fields.sort_column_number = IntegerField(required=False, default=self.sort_column_number)
        fields.sort_descending = BooleanField(required=False, default=self.sort_descending)


class PagedTable(PagedPanel):
    def __init__(self, view, page_index, columns, caption_text=None, summary=None, table_layout=None, css_id=None):
        super(PagedTable, self).__init__(view, page_index, css_id=css_id)

        def make_heading_with_sort_controls(column_number, column, view):
            heading_widget = Widget(view)
            if column.sort_key:
                heading = self.create_sorter_link(column_number, column.make_heading_widget(view))
            else:
                heading = column.make_heading_widget(view)
            heading_widget.add_child(heading)

            return heading_widget

        columns_with_sort_controls = []
        for i, column in enumerate(columns):
            make_heading_partial = functools.partial(make_heading_with_sort_controls, i, column)
            columns_with_sort_controls.append(column.with_overridden_heading_widget(make_heading_partial))

        self.table = self.add_child(Table(view, caption_text=caption_text, summary=summary))
        self.table.with_data(columns_with_sort_controls, self.current_contents)
        if table_layout:
            self.table.use_layout(table_layout)

    def create_sorter_link(self, column_number, heading_widget):
        show_control = (column_number == self.page_index.sort_column_number)
        if show_control:
            sort_descending = self.fields.toggle_sort_descending.as_input()
            link_class = 'sorted-descending' if self.page_index.sort_descending else 'sorted-ascending'
        else:
            sort_descending = self.fields.toggle_sort_descending.false_value
            link_class = None

        bookmark = Bookmark.for_widget(None,
                                       query_arguments={'sort_column_number': column_number,
                                                        'sort_descending': sort_descending})
        link = A.from_bookmark(self.view, bookmark.on_view(self.view))
        link.add_child(heading_widget)
        if link_class:
            link.append_class(link_class)
        return link

    @property
    def toggle_sort_descending(self):
        return not self.page_index.sort_descending

    @exposed
    def fields(self, fields):
        fields.toggle_sort_descending = BooleanField(writable=lambda field: False)
        
    @exposed
    def query_fields(self, fields):
        fields.update_copies(super(PagedTable, self).query_fields)
        fields.sort_column_number = self.page_index.fields.sort_column_number
        fields.sort_descending = self.page_index.fields.sort_descending


class DataTable(Div):
    """A table containing a potentially large set of data items. DataTable does not display all its items
       on the current page. It renders as a table spread over different pages between which a user can
       navigate, thus preventing a large data set sent back to a single page.

       If a :class:`DynamicColumn` is used to define the table also specifies a `sort_key`, the table
       is rendered with controls on that column heading that allows it to be sorted on that column. The
       sort operation applies to the entire dataset even though the user stays on the current page and only
       sees a subset of that data.

       :param view: (See :class:`reahl.web.fw.Widget`)
       :param columns: The :class:`DynamicColumn` instances that define the contents of the table.
       :param items: A list containing objects represented in each row of the table.
       :param css_id: (See :class:`HTMLElement`)

       :keyword items_per_page: The maximum number of rows allowed per page.
       :keyword max_page_links: The maximum number of page links to show in the page menu.
       :keyword caption_text: If given, a :class:`Caption` is added with this text.
       :keyword summary: If given, this text will be set as the summary of the contained :class:`Table` (See :class:`Table`).
       :keyword table_layout: If given, the layout is applied to the contained :class:`Table`.

    """
    def __init__(self, view, columns, items, css_id, items_per_page=10, max_page_links=5, caption_text=None, summary=None, table_layout=None):
        super(DataTable, self).__init__(view, css_id=css_id)

        self.append_class('reahl-datatable')

        self.page_index = TablePageIndex(columns, items, items_per_page=items_per_page, max_page_links=max_page_links)

        paged_css_id = '%s_paged' % css_id
        self.paged_contents = PagedTable(view, self.page_index, columns, caption_text=caption_text, summary=summary, 
                                         table_layout=table_layout, css_id=paged_css_id)
        self.page_menu = PageMenu(view, 'page_menu', self.page_index, self.paged_contents)
        self.add_children([self.page_menu, self.paged_contents])

    @property
    def table(self):
        return self.paged_contents.table





# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

This module houses stylised versions of very basic user interface
elements -- user interface elements that have a one-to-one
correspondence to HTML elements (or are of similar simplicity).



"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.component.exceptions import ProgrammerError
from reahl.component.i18n import Catalogue

import reahl.web.ui
from reahl.web.ui import A, Article, Body, Br, Div, Footer, H, Head, Header, Img, \
    Li, Link, LiteralHTML, Meta, Nav, Ol, OptGroup, P, Small, RunningOnBadge, Slot, Span, TextNode, \
    Title, Ul, WrappedInput, FieldSet, Legend, HTMLAttributeValueOption, Widget, HTMLElement, HTMLWidget

from reahl.web.bootstrap.grid import Container, ColumnLayout, ResponsiveSize


_ = Catalogue('reahl-web')


class HTML5Page(reahl.web.ui.HTML5Page):
    """A web page that may be used as the page of a web application. It ensures that everything needed by
       the framework (linked CSS and JavaScript, etc) is available on such a page.

       .. admonition:: Styling
       
          Renders as an HTML5 page with customised <head> and an empty <body>.
       
       :param view: (See :class:`reahl.web.fw.Widget`)
       :keyword title: (See :class:`reahl.web.ui.HTML5Page`)
       :keyword css_id: (See :class:`reahl.web.ui.HTMLElement`)
       
    """
    def check_form_related_programmer_errors(self):
        super(HTML5Page, self).check_form_related_programmer_errors()
        self.check_grids_nesting()
        
    def check_grids_nesting(self):
        for widget, parents_set in self.parent_widget_pairs(set([])):
            if isinstance(widget.layout, ColumnLayout):
                if not any(isinstance(parent.layout, Container) for parent in parents_set):
                    raise ProgrammerError(('%s does not have a parent with Layout of type %s.' % (widget, Container))+\
                      ' %s has a ColumnLayout, and thus needs to have an anchestor with a Container Layout.' % widget)


class Alert(Div):
    """A message box meant to alert the user of some message.

    See also: :class:`Badge`.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param message_or_widget: Message text or a Widget to display inside the Alert.
    :param severity: One of 'primary', 'secondary', 'success', 'info', 'warning', 'danger', 'light', 'dark'
        to indicate the color scheme for the Alert.

    """
    def __init__(self, view, message_or_widget, severity):
        super(Alert, self).__init__(view)
        severity_option = HTMLAttributeValueOption(severity, severity, prefix='alert', 
                                                   constrain_value_to=['primary', 'secondary',
                                                                       'success', 'info', 'warning', 'danger',
                                                                       'light', 'dark'])
        child_widget = message_or_widget
        if isinstance(message_or_widget, six.string_types):
            child_widget = TextNode(view, message_or_widget)
        self.add_child(child_widget)
        self.append_class('alert')
        self.append_class(severity_option.as_html_snippet())
        self.set_attribute('role', 'alert')


class Badge(Span):
    """An inline, visually highlighted message.

    See also: :class:`Alert`.

    :param view: (See :class:`reahl.web.fw.Widget`)
    :param message: The message for the badge.
    :param level: One of 'primary', 'secondary', 'success', 'info', 'warning', 'danger', 'light', 'dark' to indicate
                  the color scheme to be used for the Badge.
    :keyword pill: Defaults to False. If set to True the badge will look like a pill.
    """
    def __init__(self, view, message, level, pill=False):
        super(Badge, self).__init__(view)
        severity_option = HTMLAttributeValueOption(level, level, prefix='badge',
                                                   constrain_value_to=['primary', 'secondary',
                                                                       'success', 'info', 'warning', 'danger',
                                                                       'light', 'dark'])
        self.add_child(TextNode(view, message))
        self.append_class('badge')
        if pill:
            self.append_class('badge-pill')
        self.append_class(severity_option.as_html_snippet())



# Copyright 2013-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.tofu import Fixture, expected, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures

from reahl.component.exceptions import ProgrammerError

from reahl.webdev.tools import Browser, XPath

from reahl.component.modelinterface import Field, exposed, IntegerField
from reahl.web.fw import Bookmark, Widget
from reahl.web.ui import A, P, Form, TextInput, Div

from reahl.web_dev.fixtures import WebFixture


@uses(web_fixture=WebFixture)
class QueryStringFixture(Fixture):

    def is_state_labelled_now(self, label, state):
        return self.web_fixture.driver_browser.is_element_present(XPath.paragraph_containing('%s is now %s' % (label, state)))

    def is_state_now(self, state):
        return self.is_state_labelled_now('My state', state)

    def change_fragment(self, fragment):
        self.web_fixture.driver_browser.execute_script('return (window.location.hash="%s")' % fragment)

    def new_FancyWidget(self):
        fixture = self
        class MyFancyWidget(Div):
            def __init__(self, view):
                super(MyFancyWidget, self).__init__(view, css_id='sedrick')
                self.enable_refresh()
                self.add_child(P(self.view, text='My state is now %s' % self.fancy_state))
                fixture.widget = self

            @exposed
            def query_fields(self, fields):
                fields.fancy_state = IntegerField(required=False, default=1)


        return MyFancyWidget

    def new_wsgi_app(self, widget_factory=None):
        # xxxxx
        widget_factory = widget_factory or self.FancyWidget.factory()
        return self.web_fixture.new_wsgi_app(enable_js=True, child_factory=widget_factory)


@with_fixtures(WebFixture, QueryStringFixture)
def test_query_string_widget_arguments(web_fixture, query_string_fixture):
    """Widgets can have arguments that are read from a query string"""

    fixture = query_string_fixture

    class WidgetWithQueryArguments(Widget):
        def __init__(self, view):
            super(WidgetWithQueryArguments, self).__init__(view)
            self.add_child(P(view, text=self.arg_directly_on_widget))

        @exposed
        def query_fields(self, fields):
            fields.arg_directly_on_widget = Field()

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=WidgetWithQueryArguments.factory())
    browser = Browser(wsgi_app)

    browser.open('/?arg_directly_on_widget=supercalafragalisticxpelidocious')
    assert browser.lxml_html.xpath('//p')[0].text == 'supercalafragalisticxpelidocious'


@with_fixtures(WebFixture)
def test_query_string_prepopulates_form(web_fixture):
    """Widget query string arguments can be used on forms to pre-populate inputs based on the query string."""


    class ModelObject(object):
        @exposed
        def fields(self, fields):
            fields.arg_on_other_object = Field()

    class FormWithQueryArguments(Form):
        def __init__(self, view):
            self.model_object = ModelObject()
            super(FormWithQueryArguments, self).__init__(view, 'name')
            self.add_child(TextInput(self, self.model_object.fields.arg_on_other_object))

        @exposed
        def query_fields(self, fields):
            fields.arg_on_other_object = self.model_object.fields.arg_on_other_object

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=FormWithQueryArguments.factory())
    browser = Browser(wsgi_app)

    browser.open('/?arg_on_other_object=metoo')
    assert browser.lxml_html.xpath('//input')[0].value == 'metoo'


@with_fixtures(WebFixture, QueryStringFixture)
def test_widgets_with_bookmarkable_state(web_fixture, query_string_fixture):
    """If a widget has query_fields, call `.enable_refresh()` on it to let it change
       its contents without reloading the whole page,
       depending on a variable in the fragment of the current URL. This makes
       such Widgets bookmarkable by a browser."""

    fixture = query_string_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.FancyWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    # Case: the default
    assert web_fixture.driver_browser.wait_for(fixture.is_state_now, 1)
    assert fixture.widget.fancy_state == 1

    # Case: change without page load
    fixture.change_fragment('#fancy_state=2')
    assert web_fixture.driver_browser.wait_for(fixture.is_state_now, 2)
    assert fixture.widget.fancy_state == 2

    # Case: unrelated fragment changes do not trigger a reload of the widget
    previous_widget = fixture.widget
    fixture.change_fragment('#fancy_state=2&other_var=other_value')
    assert web_fixture.driver_browser.wait_for(fixture.is_state_now, 2)
    assert fixture.widget is previous_widget


@with_fixtures(WebFixture)
def test_css_id_is_mandatory(web_fixture):
    """If a Widget is enabled to to be refreshed, it must also have a css_id set."""

    class MyFancyWidget(Div):
        def __init__(self, view):
            super(MyFancyWidget, self).__init__(view)
            self.enable_refresh()

    with expected(ProgrammerError):
        MyFancyWidget(web_fixture.view)


class PartialRefreshFixture(QueryStringFixture):
    def new_FancyWidget(self):
        fixture = self
        class MyFancyWidget(Div):
            def __init__(self, view):
                super(MyFancyWidget, self).__init__(view, css_id='sedrick')
                self.enable_refresh(self.query_fields.refreshing_state)
                self.add_child(P(self.view, text='My refreshing state is now %s' % self.refreshing_state))
                self.add_child(P(self.view, text='My non-refreshing state is now %s' % self.non_refreshing_state))
                fixture.widget = self

            @exposed
            def query_fields(self, fields):
                fields.refreshing_state = IntegerField(required=False, default=1)
                fields.non_refreshing_state = IntegerField(required=False, default=2)

        return MyFancyWidget


@with_fixtures(WebFixture, PartialRefreshFixture)
def test_refreshing_only_for_specific_args(web_fixture, partial_refresh_fixture):
    """Calling `.enable_refresh()` only with specific query_fields has the effect that
       the Widget is only refreshed automatically for the particular fields passed, not
       for any of its query_fields.
    """

    fixture = partial_refresh_fixture

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=fixture.FancyWidget.factory())
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')

    # Case: the default
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My refreshing state', 1)
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My non-refreshing state', 2)

    # Case: changing the specified field only
    fixture.change_fragment('#refreshing_state=3')
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My refreshing state', 3)
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My non-refreshing state', 2)

    # Case: changing the other field only
    fixture.change_fragment('#non_refreshing_state=4')
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My refreshing state', 3)
    assert web_fixture.driver_browser.wait_for(fixture.is_state_labelled_now, 'My non-refreshing state', 2)


@with_fixtures(WebFixture)
def test_bookmarks_support_such_fragments(web_fixture):
    """Page-internal bookmarks support such bookmarkable widgets.

       These Bookmarks usually do not affect an URL - they just set widget states in different ways,
       depending on whether the client has javascript support or not.
       However, if a page was opened using the widget arguments on the querystring, a bookmark that would
       normally only have changed that argument on the hash will point to a new
       url on which the argument has been removed from the querystring and changed on the hash.
    """

    fixture = web_fixture

    internal_bookmark = Bookmark.for_widget('an ajax bookmark', query_arguments={'fancy_state':2})
    normal_bookmark = Bookmark('/', '', 'a normal bookmark')

    # You can query whether a bookmark is page_internal or not
    assert internal_bookmark.is_page_internal
    assert not normal_bookmark.is_page_internal

    # page-internal bookmarks must be added to normal ones to be useful
    usable_bookmark = normal_bookmark+internal_bookmark

    wsgi_app = web_fixture.new_wsgi_app(enable_js=True, child_factory=A.factory_from_bookmark(usable_bookmark))

    # Case: when rendered without javascript
    browser = Browser(wsgi_app)
    browser.open('/')

    a = browser.lxml_html.xpath('//a')[0]
    assert a.attrib['href'] == '/?fancy_state=2'
    assert a.text == 'an ajax bookmark'

    # Case: when rendered in a browser with javascript support
    web_fixture.reahl_server.set_app(wsgi_app)
    web_fixture.driver_browser.open('/')
    assert     web_fixture.driver_browser.is_element_present("//a[@href='/#fancy_state=2']")
    assert not web_fixture.driver_browser.is_element_present("//a[@href='/?fancy_state=2']")

    # Case: when the argument was given on the query string of the current page
    fixture.driver_browser.open('/?fancy_state=4')
    assert     web_fixture.driver_browser.is_element_present("//a[@href='/#fancy_state=2']")


class RightsScenarios(Fixture):
    def allowed(self):
        return True
    def not_allowed(self):
        return False

    @scenario
    def all_allowed(self):
        self.internal_readable = self.allowed
        self.normal_readable = self.allowed
        self.expected_readable = True
        self.internal_writable = self.allowed
        self.normal_writable = self.allowed
        self.expected_writable = True

    @scenario
    def internal_disallowed(self):
        self.internal_readable = self.not_allowed
        self.normal_readable = self.allowed
        self.expected_readable = False
        self.internal_writable = self.not_allowed
        self.normal_writable = self.allowed
        self.expected_writable = False

    @scenario
    def normal_disallowed(self):
        self.internal_readable = self.allowed
        self.normal_readable = self.not_allowed
        self.expected_readable = False
        self.internal_writable = self.allowed
        self.normal_writable = self.not_allowed
        self.expected_writable = False


@with_fixtures(RightsScenarios)
def test_bookmark_rights_when_adding(rights_scenarios):
    """When adding two Bookmarks, access rights of both are taken into account.

    """
    fixture = rights_scenarios

    internal_bookmark = Bookmark.for_widget('an ajax bookmark',
                                            query_arguments={'fancy_state':2},
                                            read_check=fixture.internal_readable,
                                            write_check=fixture.internal_writable)
    normal_bookmark = Bookmark('/', '', 'a normal bookmark',
                               read_check=fixture.normal_readable,
                               write_check=fixture.normal_writable)

    usable_bookmark = normal_bookmark+internal_bookmark
    assert usable_bookmark.read_check() is fixture.expected_readable
    assert usable_bookmark.write_check() is fixture.expected_writable



# Copyright 2015-2018 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

from reahl.tofu import Fixture, expected, scenario, uses
from reahl.tofu.pytestsupport import with_fixtures
from reahl.stubble import stubclass, EmptyStub

from reahl.webdev.tools import Browser

from reahl.component.exceptions import IsInstance, ProgrammerError
from reahl.web.fw import UserInterface, Layout
from reahl.web.ui import P, HTML5Page, Div, Header, Footer
from reahl.web.layout import PageLayout

from reahl.web_dev.fixtures import WebFixture


@with_fixtures(WebFixture)
def test_widget_layout(web_fixture):
    """A Layout is used to add children to the Widget in customised ways, and to customise the Widget itself upon construction."""

    class MyLayout(Layout):
        def customise_widget(self):
            self.widget.append_class('class-added-by-custom-layout')

        def add_wrapped(self, child):
            wrapper = self.widget.add_child(Div(self.view))
            wrapper.add_child(child)
            return child

    fixture = web_fixture

    widget_with_layout = Div(fixture.view)

    assert not widget_with_layout.has_attribute('class')
    assert not widget_with_layout.children

    widget_with_layout.use_layout(MyLayout())

    assert widget_with_layout.get_attribute('class') == 'class-added-by-custom-layout'
    assert not widget_with_layout.children

    widget_to_add = P(fixture.view)
    widget_with_layout.layout.add_wrapped(widget_to_add)

    [wrapper] = widget_with_layout.children
    assert wrapper.children == [widget_to_add]

    
@with_fixtures(WebFixture)
def test_widget_layout_errors(web_fixture):
    """A Layout can only be used with a single Widget, and a Widget can only have a single Layout."""

    fixture = web_fixture

    widget_with_layout = Div(fixture.view).use_layout(Layout())

    with expected(ProgrammerError):
        widget_with_layout.use_layout(Layout())

    re_used_layout = Layout()
    widget_with_reused_layout = Div(fixture.view).use_layout(re_used_layout)
    with expected(ProgrammerError):
        Div(fixture.view).use_layout(re_used_layout)


@uses(web_fixture=WebFixture)
class WidgetCreationScenarios(Fixture):

    @property
    def view(self):
        return self.web_fixture.view

    def new_layout(self):
        class MyLayout(Layout):
            def customise_widget(self):
                self.widget.add_child(P(self.view, text='This widget is using Mylayout'))
        return MyLayout()

    @scenario
    def use_layout_with_factory_class_method(self):
        fixture = self
        class MainUI(UserInterface):
            def assemble(self):
                self.define_view('/', title='Hello', page=HTML5Page.factory().use_layout(fixture.layout))
        self.MainUI = MainUI
        
    @scenario
    def use_layout_with_define_page(self):
        fixture = self
        class MainUI(UserInterface):
            def assemble(self):
                self.define_page(HTML5Page).use_layout(fixture.layout)
                self.define_view('/', title='Hello')
        self.MainUI = MainUI


@with_fixtures(WebFixture, WidgetCreationScenarios)
def test_widget_factory_creates_widget_with_layout(web_fixture, widget_creation_scenarios):
    """A Layout can be specified to any WidgetFactory or to UserInterface.define_page"""

    class MyLayout(Layout):
        def customise_widget(self):
            self.widget.add_child(P(self.view, text='This widget is using Mylayout'))

    layout_for_widget = MyLayout()

    class MainUI(UserInterface):
        def assemble(self):
            self.define_view('/', title='Hello', page=HTML5Page.factory(use_layout=layout_for_widget))

    fixture = widget_creation_scenarios

    wsgi_app = web_fixture.new_wsgi_app(site_root=fixture.MainUI)
    browser = Browser(wsgi_app)

    browser.open('/')
    [p] = browser.lxml_html.xpath('//p')
    assert p.text == 'This widget is using Mylayout'


@with_fixtures(WebFixture)
def test_page_layout_basics(web_fixture):
    """A PageLayout adds a Div to the body of its page (the page's document), containing a header, footer 
       with a div inbetween the two for page contents."""

    fixture = web_fixture

    layout = PageLayout()
    widget = HTML5Page(fixture.view).use_layout(layout)

    assert [layout.document] == widget.body.children[:-1]
    header, contents_div, footer = layout.document.children

    assert isinstance(header, Header)
    assert isinstance(contents_div, Div)
    assert isinstance(footer, Footer)


@with_fixtures(WebFixture)
def test_header_and_footer_slots(web_fixture):
    """PageLayout adds a Slot for Header and Footer."""

    fixture = web_fixture

    page = HTML5Page(fixture.view).use_layout(PageLayout())

    header, contents_div, footer = page.layout.document.children

    assert 'header' in header.available_slots
    assert 'footer' in footer.available_slots


@with_fixtures(WebFixture)
def test_page_layout_content_layout(web_fixture):
    """A PageLayout can be given a Layout it should use to lay out its contents Div."""

    fixture = web_fixture

    contents_layout = Layout()
    widget = HTML5Page(fixture.view).use_layout(PageLayout(contents_layout=contents_layout))

    assert widget.layout.contents.layout is contents_layout


@with_fixtures(WebFixture)
def test_page_layout_only_meant_for_html5page(web_fixture):
    """When an attempting to use a PageLayout on something other than an HTML5Page, a useful exception is raised."""

    fixture = web_fixture

    with expected(IsInstance):
        Div(fixture.view).use_layout(PageLayout())


@with_fixtures(WebFixture)
def test_page_layout_convenience_features(web_fixture):
    """A PageLayout exposes useful methods to get to its contents, and adds ids to certain elements for convenience in CSS."""

    fixture = web_fixture

    layout = PageLayout()
    widget = HTML5Page(fixture.view).use_layout(layout)
    header, contents_div, footer = layout.document.children

    assert layout.document.css_id == 'doc'
    assert header.css_id == 'hd'
    assert footer.css_id == 'ft'
    assert contents_div.css_id == 'bd'
    assert contents_div.get_attribute('role') == 'main'

    assert layout.header is header
    assert layout.contents is contents_div
    assert layout.footer is footer
